#!/usr/bin/perl -w
use strict;
use Getopt::Long;


# This script generates a DC model for an arbitrary graph, where
# vertices are cryptographers and an edge between cryptographers means that they
# share a coin.
#
# The graph is given by its nxm adjacency matrix $adj, where n is the
# number of cryptographers and m the number of coins (edges).
# If the coin k is shared between cryptographer i and j then the
# value of the (i,k) and (j,k) cell of the matrix should be 1.
#
# The methods chain(n), cycle(n), clique(n) provide an easy way to create
# standard graphs of n vertices. The "classical" DC is given for cycle(3)

my $adj = cycle(3);

#my $adj = clique(4);
#my $adj = chain(4);

#a star graph
#my $adj = [
#	[ 1, 1, 1, 1 ],
#	[ 1, 0, 0, 0 ],
#	[ 0, 1, 0, 0 ],
#	[ 0, 0, 1, 0 ],
#	[ 0, 0, 0, 1 ],
#];



my ($file, $fixorder) = (undef, undef);
GetOptions(
	"file=s"	=> \$file,
	"fixorder"	=> \$fixorder
)	or exit;

unless($file) {
	print STDERR "Usage: perl genmodel.pl --file=<filename>\n";
	exit;
}


# get the size of the matrix
#
my $n = @$adj;
my $m = @{$adj->[0]};

open F, "> $file.nm"	or die "cannot open $file.nm";

# --- Coins ---
#
my $orderCond = $fixorder ? " & pays1 != 2 " : "";
print F qq{
mdp
const double p;

module Coin1
	// coin value, 2 means not tossed
	c1 : [0..2] init 2;

	[] c1 = 2 & usersel $orderCond ->
			p/100:(c1' = 0) +
			1-p/100:(c1' = 1);
endmodule
};
for my $j (2..$m) {
	my $orderCond = $fixorder ? ", pays1=c".($j-1) : "";
	print F "module Coin$j = Coin1 [ c1=c$j $orderCond ] endmodule\n";
}
print F "\n";

# --- Master ---
#
print F q{
module Master
	usersel : bool init false;
};
for my $i (1..$n) {
	print F "\tpays$i : [0..2] init 2;\n";
}
print F "\n";
for my $i (1..$n) {
	print F "\t[] pays1 = 2 -> " .
			join(' & ', map "(pays$_' = ". ($_==$i ? 1 : 0) .")", 1..$n) .
			";\n";
}
print F q{
	[] pays1 != 2 & !usersel -> (usersel' = true);
endmodule

};

# --- Cryptographers ---
#
for my $i (1..$n) {
	my @ind = grep $adj->[$i-1][$_-1], 1..$m;

	print F
		"module Crypt$i\n" .
    		"\tout$i : [0..2] init 2;\n";

	# if $fixorder, schedule the cryptographers in their order,
	# after all coins have finished.
	# Otherwise only require the visible coins to be tossed
	#
	my $cond = $fixorder
		? ($i==1 ? "c$m" : "out".($i-1)) . " != 2 "
		: join(' & ', map "c$_ != 2", @ind);

	print F "\t[] out$i = 2 & pays$i != 2 " .
			" & $cond " .
			" ->\n";

	print F	"\t\t(out$i' = func(mod, pays$i + " .
								join(' + ', map "c$_", @ind) .
								", 2));\n";
	print F "endmodule\n";
}

close F;

open F, "> $file.pctl"	or die "cannot open $file.pctl";

# --- Observables ---
#
my $bin = 1;
for my $o (1..2**($n-1)) {
	print F "label \"observ$o\" = " .
			join(' & ', map "out$_ = ".($bin & 2**($_-1) ? 1 : 0), 1..$n) .
			";\n";
	
	my $s;
	do {
		$bin ++;
		$s = 0;
		$bin & 2**($_-1) and $s++
			for 1..$n;
	} until $s % 2 == 1;
}
print F "\n";

# --- Users ---
#
for my $i (1..$n) {
	print F "label \"user$i\" = pays$i = 1 & !usersel;\n";
}
print F "\n";

# --- Conditional probabilities
#

for my $i (1..$n) {
	for my $o (1..2**($n-1)) {
		print F "Pmax=? [ true U \"observ$o\" { \"user$i\" } ]\n";
	}
	print F "\n";
}


# --- Some standard graphs
#
sub cycle {
	my ($n) = @_;
	my $res = [];
	for my $i (0..$n-1) {
		$res->[$i] = [map +($_ == $i || $_ == ($i+1)%$n ? 1 : 0), 0..$n-1];
	}
	return $res;
}

sub chain {
	my ($n) = @_;
	my $res = [];

	$res->[0] = [1, (0)x($n-2)];
	for my $i (1..$n-2) {
		$res->[$i] = [map +($_ == $i-1 || $_ == $i ? 1 : 0), 0..$n-2];
	}
	$res->[$n-1] = [(0)x($n-2), 1];

	return $res;
}

sub clique {
	my ($n) = @_;
	my $edges = $n*($n-1)/2;
	my $res = [ map [(0)x$edges], 1..$n ];

	my $o = 0;
	for my $i (0..$n-1) {
		for my $j ($i+1..$n-1) {
			$res->[$i][$o] = 1;
			$res->[$j][$o] = 1;
			$o++;
		}
	}
	return $res;
}

